# Test stochastic dynamics generates believable event traces
#
# Copyright (C) 2017--2019 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from __future__ import print_function
import epyc
from epydemic import *
import networkx
import unittest
import six

class DummyLocus(Locus):
    def __init__(self, name):
        super(DummyLocus, self).__init__(name)

    def draw(self):
        return None

    def __len__(self):
        return 1


class FixedRatesCounter(Process):
    def __init__(self):
        super(FixedRatesCounter, self).__init__()

    def build( self, params ):
        ecr = params['eventCountRate']
        self._eventCount = [ 0 ] * len(ecr)
        for i in range(len(ecr)):
            l = self.addLocus(i, DummyLocus(i))
            self.addFixedRateEvent(l, ecr[i], lambda t, e: self.happened(i, t, e))

    def happened( self, i, t, e ):
        self._eventCount[i] = self._eventCount[i] + 1

    def results(self):
        rc = dict()
        rc['eventCount'] = self._eventCount
        return rc


class StochasticRatesTest(unittest.TestCase):

    def setUp(self):
        self._dyn = StochasticDynamics(FixedRatesCounter(), networkx.erdos_renyi_graph(100, 0.01))
        self._maxTime = 5000
        self._dyn.process().setMaximumTime(self._maxTime)

    def _checkRates(self, rc):
        eps = self._maxTime * 0.01                                # allow 1% deviation
        ecr = rc[epyc.Experiment.PARAMETERS]['eventCountRate']
        ecs = rc[epyc.Experiment.RESULTS]['eventCount']
        for i in range(0, len(ecs)):
            if ecr[i] == 0:
                self.assertEqual(ecs[i], 0)
            else:
                self.assertAlmostEqual(((ecs[i] + 0.0) / ecr[i]) / self._maxTime, 1.0, delta = eps)

    def testSameRates(self):
        '''Test same rates are generated in the correct proportions.'''
        self._checkRates(self._dyn.set(dict(eventCountRate = [ 1, 1, 1 ])).run())

    def testDifferentRates(self):
        '''Test different rates are generated in the correct proportions.'''
        self._checkRates(self._dyn.set(dict(eventCountRate = [ 1, 2, 5 ])).run())

    def testDifferentRatesOrder(self):
        '''Test different rates are generated in the correct proportions when presented in a different order.'''
        self._checkRates(self._dyn.set(dict(eventCountRate = [ 5, 2, 1 ])).run())

    def testRadicallyDifferentRates(self):
        '''Test very different rates are generated in the correct proportions.'''
        self._checkRates(self._dyn.set(dict(eventCountRate = [ 1, 2, 5, 1000 ])).run())

    def testZeroRates(self):
        '''Test that a zero-rate event isn't generated.'''
        self._checkRates(self._dyn.set(dict(eventCountRate = [ 1, 0, 1 ])).run())

    def testAllZeroRates(self):
        '''Test that nothing happens when rates are all zero.'''
        self._checkRates(self._dyn.set(dict(eventCountRate = [ 0, 0, 0 ])).run())

    def testOneRates(self):
        '''Test that things still work when there's only one event type.'''
        self._checkRates(self._dyn.set(dict(eventCountRate=[ 1 ])).run())

    def testLotsOfRates(self):
        '''Test that things still work when there are a lot of low-rate events.'''
        self._checkRates(self._dyn.set(dict(eventCountRate = [ 1 ] * 50)).run())



# SIRS as a compartmented model
#
# Copyright (C) 2017--2019 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *


class SIRS(SIR):
    '''The Susceptible-Infected-Removed-Susceptible :term:`compartmented model of disease`.
    Susceptible nodes are infected by infected neighbours, and recover to removed where
    they remain for a time before returning to susceptible models the situation in which
    disease exposure confers partial immunity in time, rather than full (SIR) or no (SIS)
    immunity.'''

    # the extra model parameter
    P_RESUSCEPT = 'pResuscept' #: Parameter for probability of losing immunity

    def __init__(self):
        super(SIRS, self).__init__()

    def build(self, params):
        '''Build the SIRS model.

        :param params: the model parameters'''

        # build SIR
        super(SIRS, self).build(params)

        # add components needed for SIRS
        pResuscept = params[self.P_RESUSCEPT]
        self.trackNodesInCompartment(self.REMOVED)

        self.addEventPerElement(self.REMOVED, pResuscept, self.resuscept)

    def resuscept(self, t, n):
        '''Perform a re-susceptibility event. This changes the compartment of
        the node from :attr:`REMOVED` to :attr:`SUSCEPTIBLE`.

        :param t: the simulation time (unused)
        :param n: the node'''
        self.changeCompartment(n, self.SUSCEPTIBLE)






# Loci for dynamics epidemic model
#
# Copyright (C) 2017--2019 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import random

class Locus(object):
    '''The locus of dynamics, where changes can happen. Loci
    are filled with nodes or edges from a network and represent some (possibly
    complete) sub-set of the elements, typically being populated and re-populated as
    a process over the network evolves. A locus is basically a set with
    some additional methods to allow or more complex behaviours, including
    customisable drawing random selection of elements.

    :param name: the locus name'''

    def __init__( self, name ):
        super(Locus, self).__init__()
        self._name = name
        self._elements = set()

    def name( self ):
        '''Returns the name of the locus.

        :returns: the locus' name'''
        return self._name

    def __len__( self ):
        '''Return the number of elements at the locus.

        :returns: the number of elements'''
        return len(self.elements())

    def elements( self ):
        '''Return the underlying elements of the locus.

        :returns: the elements'''
        return self._elements
    
    def draw(self):
        '''Draw a random element from the locus. The default performs a simple
        draw that is equiprobable across all the elements. The locus remains unchanged:
        drawing simply selects ans returns an element at random.

        :returns: a random element at the locus'''
        if len(self) == 0:
            raise ValueError('Trying to draw element from empty locus {n}'.format(n = self.name()))
        e = (random.sample(self.elements(), 1))[0]
        return e

    # ---------- Handlers ----------

    def addHandler(self, g, e):
        '''Handler called when an element is added to the network. This is used to indicate
        the the population of nodes or edges has changed.

        :param g: the network
        :param e: the element'''
        self._elements.add(e)

    def leaveHandler( self, g, e ):
        '''Handler for when an element leaves the locus due to changes in circumstances,
        not changes in population.

        :param g: the network
        :param e: the element'''
        self._elements.discard(e)

    def enterHandler( self, g, e ):
        '''Handler for when an element enters the locus due to changes in circumstances,
        not changes in population. 

        :param g: the network
        :param e: the element'''
        self._elements.add(e)

    def removeHandler(self, g, e):
        '''Handler called when an element is removed from the network.  This is used to indicate
        the the population of nodes or edges has changed.

        :param g: the network
        :param e: the element'''
        self._elements.discard(e)




# Test the behaviour of the addition-deletion process
#
# Copyright (C) 2017--2019 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from __future__ import print_function
from epydemic import *
from test.compartmenteddynamics import CompartmentedDynamicsTest
import epyc
import unittest
import networkx

class AddDeleteRecorder(AddDelete):

    N = 'networkSize'

    def __init__(self):
        super(AddDeleteRecorder, self).__init__()

    def results( self ):
        '''Save the size of the resulting network.

        :returns: a dict of experimental results'''
        rc = super(AddDeleteRecorder, self).results()
        rc[self.N] = self.network().order()
        return rc


class AddDeleteTest(unittest.TestCase):

    def setUp(self):
        '''Set up the experimental parameters and process.'''
        N = 5000
        kmean = 10
        phi = (kmean + 0.0) / N
        self._network = networkx.erdos_renyi_graph(N, phi)
        self._maxTime = 5000

        self._params = dict()
        self._params[AddDelete.DEGREE] = 10
        self._process = AddDeleteRecorder()
        self._process.setMaximumTime(self._maxTime)
        self._e = StochasticDynamics(self._process, self._network)

    def testRun(self):
        '''Test that the process runs and adds and deletes at roughly equal rates.'''
        self._params[AddDelete.P_ADD] = 1
        self._params[AddDelete.P_DELETE] = 1
        rc = self._e.set(self._params).run()
        self.assertAlmostEqual(rc[epyc.Experiment.RESULTS][AddDeleteRecorder.N], self._network.order(), delta = int((self._network.order() + 0.0) * 0.1))

    def testRunFaster(self):
        '''Test that the process runs and adds faster than it deletes.'''
        self._params[AddDelete.P_ADD] = 1
        self._params[AddDelete.P_DELETE] = 0.5
        rc = self._e.set(self._params).run()
        dn = self._maxTime * (self._params[AddDelete.P_ADD] - self._params[AddDelete.P_DELETE])
        print(rc[epyc.Experiment.RESULTS][AddDeleteRecorder.N], self._network.order(), dn)
        self.assertAlmostEqual(rc[epyc.Experiment.RESULTS][AddDeleteRecorder.N], self._network.order() + dn, delta = int((self._network.order() + 0.0) * 2 * 0.1))



# Test SIS under different dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
from test.compartmenteddynamics import CompartmentedDynamicsTest

import epyc
import unittest
import networkx

class SISTest(unittest.TestCase, CompartmentedDynamicsTest):

    def setUp( self ):
        '''Set up the experimental parameters and experiment.'''
        
        # single experiment
        self._params = dict()
        self._params[SIS.P_INFECT] = 0.1
        self._params[SIS.P_INFECTED] = 0.01
        self._params[SIS.P_RECOVER] = 0.05
        self._network = networkx.erdos_renyi_graph(1000, 0.005)

        # lab run
        self._lab = epyc.Lab()
        self._lab[SIS.P_INFECT] = [ 0.1, 0.3 ]
        self._lab[SIS.P_INFECTED] = 0.01
        self._lab[SIS.P_RECOVER] = 0.05

        # model
        self._model = SIS()

        # maximum time needed as disease may be endemic
        self._model.setMaximumTime(200)



# Networks dynamics simulation base class
#
# Copyright (C) 2017--2019 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import epyc
import networkx


class Dynamics(epyc.Experiment, object):
    '''An abstract simulation framework for running a process over a network.
    This is the abstract base class
    for implementing different kinds of dynamics as computational experiments
    suitable for running under ``epyc``. Sub-classes provide synchronous and stochastic
    (Gillespie) simulation dynamics.

    The dynamics runs a network process provided as a :class:`Process`
    object. It optionally takes a prototype network over which the process runs, which
    is copied for every run. If not provided at construction, the prototype should be
    proivided by calling :meth:`setPrototypeNetwork` beefore trying to run a simulation.

    :param p: network process to run
    :param g: prototype network (optional)'''

    # Additional metadata elements
    TIME = 'simulation_time'      #: Metadata element holding the logical simulation end-time.
    EVENTS = 'simulation_events'  #: Metadata element holding the number of events that happened.

    def __init__( self, p, g = None ):
        super(Dynamics, self).__init__()
        self._graphPrototype = g                 # prototype copied for each run
        self._graph = None                       # working copy of prototype
        self._process = p                        # network process to run


    # ---------- Configuration ----------

    def network( self ):
        '''Return the network this dynamics is running over. This will return None
        unless we're actually running a simulation.

        :returns: the network'''
        return self._graph

    def setNetworkPrototype( self, g ):
        '''Set the network the dynamics will run over. This will be
        copied for each run of an individual experiment.

        :param g: the network'''
        self._graphPrototype = g

    def networkPrototype( self ):
        '''Return the prototype network used to create the working
        copy.

        :returns: the prototype network'''
        return self._graphPrototype
    
    def process(self):
        '''Return the network process being run.

        :returns: the process'''
        return self._process


    # ---------- Results ----------

    def experimentalResults(self):
        '''Report the process' experimental results. This simply calls through to
        the :meth:`Process.results` method of the process being simulated.

        :returns: the results of the process'''
        return self._process.results()


    # ---------- Set-up and tear-down ----------

    def setUp(self, params):
        '''Set up the experiment for a run. This performs the inherited actions, then
        copies the prototype network and builds the network process that the dynamics is to run.

        :params params: the experimental parameters'''
        super(Dynamics, self).setUp(params)

        # make a copy of the network prototype
        self._graph = self.networkPrototype().copy()

        # build the process
        self._process.reset()
        self._process.setNetwork(self.network())
        self._process.build(params)
        self._process.setUp(params)

    def tearDown(self):
        '''At the end of each experiment, throw away the copy and any posted by un-executed events.'''
        super(Dynamics, self).tearDown()

        # throw away the worked-on model and any posted events that weren't run
        self._graph = None
        self._posted = []


    # ---------- Pending event handling ----------

    def runPendingEvents(self, t):
        '''Retrieve and fire any pending events at time t. This method handles
        the case where firing an event posts another event that needs to be run
        before other already-posted events coming before time t: in other words,
        it ensures that the simulation order is respected.

        :param t: the current time
        :returns: the number of events fired'''
        n = 0
        while True:
            pef = self._process.nextPendingEventBefore(t)
            if pef is None:
                # no more pending events, return however many we've fired already
                return n
            else:
                # fire the event
                pef()
                n += 1




# Test compartmented model basic functions
#
# Copyright (C) 2017--2019 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import unittest
import networkx
import six

class CompartmentedModelTest(unittest.TestCase):

    def setUp( self ):
        '''Set up the experimental parameters and experiment.'''
        self._er = networkx.erdos_renyi_graph(1000, 0.005)
        self._params = dict(pInfect = 0.1,
                            pInfected = 0.01,
                            pRemove = 0.05)
       
    def testPopulation( self ):
        '''Test populating a model.'''
        e = StochasticDynamics(SIR(), self._er)
        e.setUp(self._params)

    def testDuplicateNodeLocus( self ):
        '''Test we can't duplicate node-loci names'''
        m = SIR()
        params = dict(pInfect = 0.1,
                      pInfected = 1.0,    # infect all nodes initially
                      pRemove = 0.05)
        m.build(params)
        with self.assertRaises(Exception):
            m.trackNodesInCompartment(SIR.INFECTED)      # part of SIR already

    def testDuplicateEdgeLocus( self ):
        '''Test we can't duplicate edge-loci names'''
        m = SIR()
        params = dict(pInfect = 0.1,
                      pInfected = 1.0,    # infect all nodes initially
                      pRemove = 0.05)
        m.build(params)
        with self.assertRaises(Exception):
            m.trackEdgesBetweenCompartments(SIR.SUSCEPTIBLE, SIR.INFECTED, name = SIR.SI)      # part of SIR already

    def testLocusNames(self):
        '''Test that a locus gets an automatic name.'''
        m = SIR()

        # add new loci
        m.trackNodesInCompartment(SIR.REMOVED)
        self.assertIn(SIR.REMOVED, m._loci.keys())
        m.trackEdgesBetweenCompartments(SIR.REMOVED, SIR.SUSCEPTIBLE)
        self.assertIn("{l}-{r}".format(l = SIR.REMOVED, r = SIR.SUSCEPTIBLE), m._loci.keys())

    def testLoci( self ):
        '''Test we can populate loci correctly.'''
        g = networkx.Graph()
        g.add_edges_from([(1, 2), (2, 3), (1, 4), (3, 4)])
        m = SIR()
        e = StochasticDynamics(m, g)
        params = dict(pInfect = 0.1,
                      pInfected = 1.0,    # infect all nodes initially
                      pRemove = 0.05)
        e.setUp(params)

        # keep track of the other compartments as well
        m.trackNodesInCompartment(SIR.SUSCEPTIBLE)
        m.trackNodesInCompartment(SIR.REMOVED)
        
        # all nodes in I
        six.assertCountEqual(self, m._loci[SIR.INFECTED].elements(), [1, 2, 3, 4])

        # one node from I into S, two edges into SI
        m.changeCompartment(1, SIR.SUSCEPTIBLE)
        six.assertCountEqual(self, m._loci[SIR.INFECTED].elements(), [2, 3, 4])
        six.assertCountEqual(self, m._loci[SIR.SUSCEPTIBLE].elements(), [1])
        six.assertCountEqual(self, m._loci[SIR.SI].elements(), [(1, 2), (1, 4)])

        # recover the infected node
        m.changeCompartment(1, SIR.REMOVED)
        six.assertCountEqual(self, m._loci[SIR.INFECTED].elements(), [2, 3, 4])
        six.assertCountEqual(self, m._loci[SIR.SUSCEPTIBLE].elements(), [])
        six.assertCountEqual(self, m._loci[SIR.REMOVED].elements(), [1])
        six.assertCountEqual(self, m._loci[SIR.SI].elements(), [])

    def testSkeletonise( self ):
        '''Test that a network skeletonises correctly'''
        m = SIR()
        g = networkx.Graph()
        m.setNetwork(g)
        g.add_edges_from([ (1, 2), (2, 3), (1, 4), (3, 4) ])

        # mark some edges as occupied
        m.markOccupied((1, 2), 0)
        m.markOccupied((1, 4), 0)

        # check skeletonisation
        g2 = m.skeletonise()
        six.assertCountEqual(self, g2.edges(), [(1, 2), (1, 4)])
        six.assertCountEqual(self, g2.nodes(), [ 1, 2, 3, 4])

    def testAddNode(self):
        '''Test that a node can be added and land in the right compartment.'''
        m = SIR()
        g = networkx.Graph()
        m.setNetwork(g)

        m.addNode(1, compartment = SIR.INFECTED)
        six.assertCountEqual(self, m.compartment(SIR.INFECTED), [1])

        m.addNode(2, compartment = SIR.INFECTED, h = "test")
        six.assertCountEqual(self, m.compartment(SIR.INFECTED), [1, 2])
        self.assertEqual(g.nodes[2]['h'], "test")

    def testAddNodes(self):
        '''Test that nodes can be added and land in the right compartment.'''
        m = SIR()
        g = networkx.Graph()
        m.setNetwork(g)

        ns = [1,2,3]
        m.addNodesFrom(ns, compartment = SIR.INFECTED, h = 'test')
        six.assertCountEqual(self, m.compartment(SIR.INFECTED), ns)
        for n in ns:
            self.assertEqual(g.nodes[n]['h'], "test")

    def testRemoveNode(self):
        '''Test that a node can be removed.'''
        m = SIR()
        g = networkx.Graph()
        m.setNetwork(g)

        m.addNode(1, compartment = SIR.INFECTED)
        six.assertCountEqual(self, m.compartment(SIR.INFECTED), [1])
        m.removeNode(1)
        six.assertCountEqual(self, m.compartment(SIR.INFECTED), [])

    def testAddEdge(self):
        '''Test we can add an edge, keeping all the data straight.'''
        m = SIR()
        g = networkx.Graph()
        m.setNetwork(g)
        m.trackEdgesBetweenCompartments(SIR.INFECTED, SIR.SUSCEPTIBLE, name = SIR.SI)
        m.addNode(1, compartment = SIR.INFECTED)
        m.addNode(2, compartment = SIR.SUSCEPTIBLE)
        six.assertCountEqual(self, m.compartment(SIR.INFECTED), [1])
        six.assertCountEqual(self, m.compartment(SIR.SUSCEPTIBLE), [2])
        m.addEdge(1, 2)
        self.assertEqual(len(m[SIR.SI]), 1)

    def testRemoveEdge(self):
        '''Test we can remove an edge, keeping all the data straight.'''
        m = SIR()
        g = networkx.Graph()
        m.setNetwork(g)
        m.trackEdgesBetweenCompartments(SIR.INFECTED, SIR.SUSCEPTIBLE, name = SIR.SI)
        m.addNode(1, compartment = SIR.INFECTED)
        m.addNode(2, compartment = SIR.SUSCEPTIBLE)
        m.addNode(3, compartment = SIR.SUSCEPTIBLE)
        m.addEdge(1, 2)
        m.addEdge(1, 3)
        m.addEdge(3, 2)
        self.assertEqual(len(m[SIR.SI]), 2)
        m.removeEdge(1, 2)
        self.assertEqual(len(m[SIR.SI]), 1)
        print(list(g.edges()))
        m.removeEdge(2, 3)
        print(list(g.edges()))
        self.assertEqual(len(m[SIR.SI]), 1)






# Test the SIR and SIS fixed-time recovery variants
#
# Copyright (C) 2017--2019 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from __future__ import print_function
from epydemic import *
import epyc
import unittest
import networkx
import six

class FixedRecoveryTest(unittest.TestCase):

    def setUp( self ):
        '''Set up the experimental parameters and experiment.'''
        
        # single experiment
        self._params = dict()
        self._params[SIR_FixedRecovery.P_INFECT] = 0.1
        self._params[SIR_FixedRecovery.P_INFECTED] = 0.01
        self._params[SIR_FixedRecovery.T_INFECTED] = 1
        self._network = networkx.erdos_renyi_graph(1000, 0.005)

    def testRunSingleSIRSynchronous( self ):
        '''Test a single run of a fixed-period SIR under synchronous dynamics.'''
        e = SynchronousDynamics(SIR_FixedRecovery(), self._network)
        rc = e.set(self._params).run()
        if not rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS]:
            print(rc[epyc.Experiment.METADATA][epyc.Experiment.TRACEBACK])
            self.assertTrue(rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS])
        else:
            six.assertCountEqual(self, rc[epyc.Experiment.RESULTS], ['S', 'I', 'R'])
            self.assertTrue(rc[epyc.Experiment.RESULTS]['S'] > 0)
            self.assertTrue(rc[epyc.Experiment.RESULTS]['I'] == 0)
            self.assertTrue(rc[epyc.Experiment.RESULTS]['R'] > 0)
            self.assertEqual(rc[epyc.Experiment.RESULTS]['S'] + rc[epyc.Experiment.RESULTS]['R'], self._network.order())

    def testRunSingleSIRStochastic( self ):
        '''Test a single run of a fixed-period SIR under stochastic dynamics.'''
        e = StochasticDynamics(SIR_FixedRecovery(), self._network)
        rc = e.set(self._params).run()
        if not rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS]:
            print(rc[epyc.Experiment.METADATA][epyc.Experiment.TRACEBACK])
            self.assertTrue(rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS])
        else:
            six.assertCountEqual(self, rc[epyc.Experiment.RESULTS], [ 'S', 'I', 'R' ])
            self.assertTrue(rc[epyc.Experiment.RESULTS]['S'] > 0)
            self.assertTrue(rc[epyc.Experiment.RESULTS]['I'] == 0)
            self.assertTrue(rc[epyc.Experiment.RESULTS]['R'] > 0)
            self.assertEqual(rc[epyc.Experiment.RESULTS]['S'] + rc[epyc.Experiment.RESULTS]['R'], self._network.order())
  



# Networks process base class
#
# Copyright (C) 2017--2019 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
from heapq import *
import six


class Process(object):
    """A process that runs over a network. This is the abstract base class for all network processes.
    It provides the essential routines to build, set-up, run, and extract results from a
    process. Sub-classes provide the actual behaviour by defining simulation events and
    attaching them to the network in different ways.

    A process defines is largely declarative, in the sense that it sets up event handlers
    and their probabilities. These are used by a process dynamics (a sub-class of :class:`Dynamics`)
    to actually run a simulation of the process.

    The process class includes an interface for interacting with the working network. The basic
    interface is extended and overridden in different sub-classes that interact with the network
    in different ways.
    """

    # defaults
    DEFAULT_MAX_TIME = 20000      #: Default maximum simulation time.

    def __init__(self):
        super(Process, self).__init__()

        # set the default maximum time, which persists across runs of the process
        self._maxTime = self.DEFAULT_MAX_TIME

        # reset this instance
        self.reset()


    # ---------- Setup and initialisation ----------

    def reset(self):
        """Reset the process ready to be built. This resets all the internal process state
        variables. Sub-classes should call the base method to make sure that the event
        sub-system is properly reset."""
        self._loci = dict()                          # loci for events
        self._g = None                               # working network
        self._perElementEvents = []                  # events that occur per-element
        self._perLocusEvents = []                    # events that occur per-locus
        self._postedEvents = []                      # pri-queue of fixed-time events
        self._eventId = 0                            # counter for posted events

    def build(self, params):
        """Build the process model. This must be overridden by sub-classes, and should
        create the various elements of the model.

        :param params: the model parameters"""
        raise NotImplementedError('build')

    def setUp(self, params):
        """Set up the network under the given dynamics. The default does
        nothing: sub-classes should override it to initialise node states
        or establish other properties ready for the experiment.

        :param params: the simulation parameters"""
        pass

    def setNetwork(self, g):
        """Set the network the process is running over.

        :param g: the network"""
        self._g = g

    def network(self):
        """Return the network the process is running over.

        :returns: the network"""
        return self._g

    def setMaximumTime(self, t):
        """Set the maximum default simulation time. The default is given by :attr:`DEFAULT_MAX_TIME`.
        This is used by :meth:`atEquilibrium` as the default way to determine equilibrium.

        Setting the maximum time persists across runs of the process, and isn't reset to its
        default by calling :meth:`reset`.

        :param t: the maximum simulation time"""
        self._maxTime = t

    def maximumTime(self):
        """Return the maximum assumed simulation time.

        :returns: the maximum simulation time"""
        return self._maxTime


   # ---------- Termination and results ----------

    def atEquilibrium(self, t):
        """Test whether the process is an equilibrium. The default simply checks whether
        the simulation time exceeds the maximum given by :meth:`setMaximumTime`, which
        defaults to an arbitrary value given by :attr:`DEFAULT_MAX_TIME`.
        Sub-classes may override this to provide a more sensible definition: in many
        cases it makes sense to keep calling this method as part of any more advanced
        test, so as to cut-off runaway processes.

        :param t: the current simulation time
        :returns: True if the proceess is now at equilibrium"""
        return (t >= self.maximumTime())

    def results(self):
       """Create  and return an empty dict to be filled with experimental results.
       Sub-classes should extend this method to add results to the dict.

       :returns: an empty dict for experimental results"""
       return dict()


    # ---------- Accessing and evolving the network ----------

    def addNode(self, n, **kwds):
        """Add a node to the working network. Any keyword arguments added as node attributes

        :param n: the new node
        :param kwds: (optional) node attributes"""
        self.network().add_node(n, **kwds)

    def addNodesFrom(self, ns, **kwds):
        """Add all the nodes in the given iterable to the working network. Any keyword arguments are
        added as node attributes. This works by calling :meth:`addNode` for each
        element of the iterable.

        :param ns: an iterable collection of nodes
        :param kwds: (optional) node attributes"""
        for n in ns:
            self.addNode(n, **kwds)

    def removeNode(self, n):
        """Remove a node from the working network.

        :param n: the node"""
        self.network().remove_node(n)

    def removeNodesFrom(self, ns):
        """Remove all the nodes in the given iterable froim the working network. This works by
        iteratively calling :meth:`removeNode` for each element of the iterable collection.

        :param ns: an iterable collection of nodes"""
        for n in ns:
            self.removeNode(n)

    def addEdge(self, n, m, **kwds):
        """Add an edge between nodes. If the nodes do not exist, they are created by calling
        :meth:`addNode`. Any keyword arguments are addaed as edge attributes.

        :param n: the start node
        :param m: the end node
        :param kwds: (optional) edge attributes"""
        g = self.network()
        if n not in g:
            self.addNode(n)
        if m not in g:
            self.addNode(m)
        g.add_edge(n, m, **kwds)

    def addEdgesFrom(self, es, **kwds):
        """Add all the edges in the given iterable to the working network. Any keyword arguments are
        added as node attributes. This works by calling :meth:`addEdge` for each
        element of the iterable.

        :param es: an iterable collection of edges
        :param kwds: (optional) node attributes"""
        for e in es:
            (n, m) = e
            self.addEdge(n, m, **kwds)

    def removeEdge(self, n, m):
        """Remove an edge from the working network.

        :param n: the start node
        :param m: the end node"""
        self.network().remove_edge(n, m)

    def removeEdgesFrom(self, es):
        """Remove all the edges in the given iterable collection from the working network. This works
        by iteratively calling :meth:`removeEdge` for each edge in the iterable.

        :param es: an iterable collection of edges"""
        for e in es:
            self.removeEdge(*e)


    # ---------- Loci for events ----------

    def addLocus(self, n, l=None):
        """Add a named locus.

        :param n: the locus name
        :param l: the locus (defaults to a simple set-based locus)
        :returns: the locus"""
        if n in self._loci.keys():
            raise Exception("Trying to overwrite existing locus {n}".format(n = n))
        if l is None:
            l = Locus(n)
        self._loci[n] = l
        return l

    def __setitem__(self, n, l):
        """Add a named locus. Equivalent to :meth:`addLocus`

        :param n: the locus name
        :param l: the locus
        :returns: the locus"""
        return self.addLocus(n, l)

    def locus(self, n):
        """Return the named locus.

        :param n: the locus name
        :returns: the locus"""
        return self._loci[n]

    def __getitem__(self, n):
        """Return the given locus. Equivalent to :meth:`locus`.

        :param n: the locus name
        :returns: the locus"""
        return self.locus(n)

    def __contains__(self, l):
        """Check whether the given locus is defined for the process.

        :returns: the name of the loci"""
        return l in self._loci.keys()

    def __iter__(self):
        """Return an iterator over the loci in the process.

        :returns: an iterator"""
        return iter(self._loci.keys())


    # ---------- Probabilistic events (occurring with some probability, either per-locus or per-element) ----------

    def addEventPerElement(self, l, p, ef):
        """Add a probabilistic event at a locus, occurring with a particular (fixed)
        probability for each element of the locus, and calling the :term:`event function`
        when it is selected. The locus may be a :class:`Locus` object or a string, which
        is taken to be a locus of this process.

        Unlike fixed probability events added by :meth:`addFixedRateEvent`, a per-element
        event happens with the same probability to each element in the locus. This leads to
        larger loci giving rise to a higher rate of events.

        :param l: the locus or locus name
        :param p: the event probability
        :param ef: the event function"""
        if isinstance(l, six.string_types):
            l = self.locus(l)
        self._perElementEvents.append((l, p, ef))

    def addFixedRateEvent(self, l, p, ef):
        """Add a probabilistic event at a locus, occurring with a particular (fixed)
        probability, and calling the :term:`event function`
        when it is selected. The locus may be a :class:`Locus` object or a string, which
        is taken to be the name of a locus of this process.

        Unlike fixed rate events added by :meth:`addEventPerElement`, a fixed probability
        event happens with the same probability regardless of how many elements are in
        the locus.

        :param l: the locus or locus name
        :param p: the event probability
        :param ef: the event function"""
        if isinstance(l, six.string_types):
            l = self.locus(l)
        self._perLocusEvents.append((l, p, ef))

    def perElementEventDistribution(self, t):
        """Return the distribution of per-element events at time t.

        :param t: the simulation time
        :returns: a list of (locus, probability, event function) triples"""
        return self._perElementEvents

    def fixedRateEventDistribution(self, t):
        """Return the distribution of fixed-rate events at time t.

        :param t: the simulation time
        :returns: a list of (locus, probability, event function) triples"""
        return self._perLocusEvents

    def eventRateDistribution(self, t):
        """Return the event distribution, a sequence of (l, r, f) triples
        where l is the locus where the event occurs, r is the rate at
        which an event occurs, and f is the event function called to
        make it happen.

        Note the distinction between a *rate* and a *probability*:
        the former can be obtained from the latter simply by
        multiplying the event probability by the number of times it's
        possible in the current network, which is the population
        of nodes or edges in a given state.

        It is perfectly fine for an event to have a zero rate. The process
        is assumed to have reached equilibrium if all events have zero rates.

        :param t: current time
        :returns: a list of (locus, rate, event function) triples"""
        rates = []

        # convert per-element events to rates
        for (l, p, ef) in self.perElementEventDistribution(t):
            rates.append((l, p * len(l), ef))

        # add fixed-rate events for non-empty loci
        for (l, p, ef) in self.fixedRateEventDistribution(t):
            if len(l) > 0:
                rates.append((l, p, ef))
        return rates


    # ---------- Posted events (occurring at a fixed time) ----------

    def _nextEventId(self):
        """Generate a sequence number for a posted event. This is used to ensure that
        all event triples pushed onto the priqueue are unique in their first two elements,
        and therefore never try to do comparisons with functions (the third element).

        :returns: a sequence number"""
        id = self._eventId
        self._eventId += 1
        return id

    def postEvent(self, t, e, ef):
        """Post an event to happen at time t, calling the :term:`event function`
        at time t.

        :param t: the current time
        :param e: the element (node or edge) on which the event occurs
        :param ef: the event function"""
        heappush(self._postedEvents, (t, self._nextEventId(), (lambda: ef(t, e))))

    def postRepeatingEvent(self, t, dt, e, ef):
        """Post an event that re-occurs at interval dt, taken from now.

        :param t: the start time
        :param dt: the interval
        :param e: the element (node or edge) on which the event occurs
        :param ef: the element function"""

        def repeat(tc, e):
            ef(tc, e)
            tp = tc + dt
            heappush(self._postedEvents, (tp, self._nextEventId(), (lambda: repeat(tp, e))))

        heappush(self._postedEvents, (t, self._nextEventId(), (lambda: repeat(t, e))))

    def nextPendingEventBefore(self, t):
        """Return the next pending event to occur at or before time t.

        :param t: the current time
        :returns: a pending event function or None"""
        if len(self._postedEvents) > 0:
            # we have events, grab the soonest
            (et, id, pef) = heappop(self._postedEvents)
            if et <= t:
                # event should have occurred, return it
                return pef
            else:
                # this (and therefore all further events) are in the future, put it back
                heappush(self._postedEvents, (et, id, pef))
                return None
        else:
            # we don't have any events
            return None

    def pendingEvents(self, t):
        """Retrieve any :term:`posted event` scheduled to be fired at or
        before time t. The pending events are returned in the form of
        zero-argument functions that can simply be called to fire
        the corresponding event. The events are returned as a list with the
        earliest-posted event first.

        Be aware that running the returned events in order may not be enough to
        accurately run the simulation in the case where firing an
        event causes another event to be posted before t. It may be
        easier to use :meth:`Dynamics.runPendingEvents` to run all pending
        events, which handles this case automatically.

        :param t: the current time
        :returns: a (possibly empty) list of pending event functions"""
        pending = []
        while True:
            pef = self.nextPendingEventBefore(t)
            if pef is None:
                # no more events pending, return those we've got
                return pending
            else:
                # store the pending event function
                pending.append(pef)








# Canonical addition-deletion process
#
# Copyright (C) 2017--2019 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from __future__ import print_function
from epydemic import *


class AddDelete(Process):
    '''A process to manage an addition-deletion network, in which nodes are added
    and deleted at random according to some probability. Nodes that are added
    are connected to existing nodes.

    The default behaviour has fixed addition and removal probabilities, independent
    of the size of the network, and a fixed degree for added nodes.'''

    # parameters
    P_ADD = "pAdd"  #: Parameter for the node addition probability
    P_DELETE = "pDelete"  #: Parameter for the node deletion probability
    DEGREE = "addDegree"  #: Degree of newly-added nodes

    # loci
    NODES = "allnodes"       #: Name of the locus holding all nodes in the network

    def __init__(self):
        super(AddDelete, self).__init__()


    # ---------- Setup and initialisation ----------

    def build(self, params):
        '''Build the model. This method expects parameters for the node addition
        and deletion probabilities, and for the degree of newly-created nodes.

        :param params: the model parameters'''
        pAdd = params[self.P_ADD]
        pDelete = params[self.P_DELETE]

        # stash the degree of new nodes for the events
        self._c = params[self.DEGREE]

        # keep track of all the nodes and edges
        self.addLocus(self.NODES)

        # add events occurring at constant probability regardless of the network size
        self.addFixedRateEvent(self.NODES, pAdd, self.add)
        self.addFixedRateEvent(self.NODES, pDelete, self.remove)

    def setUp(self, params):
        super(AddDelete, self).setUp(params)

        # add all nodes to the all-nodes locus
        g = self.network()
        for n in g.nodes():
            self[self.NODES].addHandler(g, n)


    # ---------- Accessing and evolving the network ----------

    def addNode(self, n, **kwds):
        '''Add a node to the working network. Any keyword arguments added as node attributes.

        :param n: the new node
        :param kwds: (optional) node attributes'''
        super(AddDelete, self).addNode(n, **kwds)
        self[self.NODES].addHandler(self.network(), n)
        #print('Added {n}'.format(n=n))

    def newNodeName(self):
        '''Generate a new name for a node to be added. This is guaranteed not to be
        the name of another node in the network (although it might possibly re-use
        a name of a node that's been removed).

        :returns: the generated name'''
        # sd: might be faster to generate a random name?
        g = self.network()
        i = g.order() + 1
        while i in g.nodes():
            i = i + 1
        return i

    def addNewNode(self, **kwds):
        '''Add a new node to the network with a new, unused name. Any keyword arguments are
        added as node attributes.

        :returns: the generated name of the new node'''
        i = self.newNodeName()
        self.addNode(i, **kwds)
        return i

    def removeNode(self, n):
        '''Remove a node from the working network.

        :param n: the node'''
        self[self.NODES].removeHandler(self.network(), n)
        super(AddDelete, self).removeNode(n)
        #print('removed {n}'.format(n=n))


    # ---------- Events ----------

    def add(self, t, e):
        '''Add a node to the network, connecting it at random to
        other nodes. The degree of the new node is given by the :attr:`DEGREE` parameter,
        with the nodes being selected at random from the entire network.

        :param t: the current simulation time (not used)
        :param e: the element (not used)'''

        # create a new node
        i = self.addNewNode()

        # link to c other nodes (not including i) with uniform probability
        ns = self[self.NODES]
        es = set()
        for m in range(self._c):
            # a probably unnecessary test for parallel edges and self-loops
            while True:
                j = ns.draw()
                if (j not in es) and (i != j):
                    break
            es.add(j)
        for j in es:
            self.addEdge(i, j)

    def remove(self, t, n):
        '''Remove a node from the network.

        :param t: the current simulation time (not used)
        :param n: the node to be removed'''
        self.removeNode(n)


# Compartmented models base class
#
# Copyright (C) 2017--2019 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from __future__ import print_function
from epydemic import *
import numpy
import collections


class CompartmentedLocus(Locus):
    '''A locus based on the compar6tments that nodes reside in.

    :param name: the locus' name'''

    def __init__(self, name):
        super(CompartmentedLocus, self).__init__(name)

    def compartments(self):
        '''Return the compartments this locus monitors. This should be
        overridden by sub-classes as required.

        :returns: the compartments'''
        return []
   

class CompartmentedNodeLocus(CompartmentedLocus):
    '''A locus for dynamics occurring at a single node, where the
    locus tracks all nodes in a single compartment.

    :param name: the locus' name
    :param c: the compartment'''

    def __init__(self, name, c):
        super(CompartmentedNodeLocus, self).__init__(name)
        self._compartment = c

    def compartments(self):
        '''Return the node compartment we monitor.

        :returns: the compartment'''
        return [ self._compartment ]


class CompartmentedEdgeLocus(CompartmentedLocus):
    '''A locus for dynamics occurring at an edge, where the locus tracks all
    edges whose endpoint nodes are in specified compartments. Since the network may
    also be adaptive, we need to track additionals and removals of edges too.

    :param name: the locus' name
    :param l: the left compartment
    :param r: the right compartment'''

    def __init__(self, name, l, r):
        super(CompartmentedEdgeLocus, self).__init__(name)
        self._left = l
        self._right = r

    def compartments(self):
        '''Return the compartments of the node endpoints we monitor.

        :returns: the compartments'''
        return [ self._left, self._right ]

    def matches(self, g, n, m):
        '''Test whether the given edge has the right compartment endpoints for this compartment. The
        method returns 1 if the edge has the right compartments in the orientation (n, m), -1 if it has
        the right compartments in orientation (m, n), and 0 otherwise.

        :param g: the network
        :param n: the first node
        :param m: the second node
        :returns: match status -1, 0, or 1'''
        if (g.nodes[n][CompartmentedModel.COMPARTMENT] == self._right) and (g.nodes[m][CompartmentedModel.COMPARTMENT] == self._left):
            return -1
        else:
            if (g.nodes[n][CompartmentedModel.COMPARTMENT] == self._left) and (g.nodes[m][CompartmentedModel.COMPARTMENT] == self._right):
                return 1
            else:
                return 0

    def addHandler(self, g, e):
        '''An edge is added to the network, check if its endpoint compartments match the
        locus and add it if so. 

        :param, g: the network
        :param e: the edge'''
        (n, m) = e
        es = self.elements()
        match = self.matches(g, n, m)
        if match == -1:
            #print('edge ({m}, {n}) added to {l}'.format(n = n, m = m, l = self._name))
            es.add((m, n))
        else:
            if match == 1:
                #print('edge ({n}, {m}) added {l}'.format(n = n, m = m, l = self._name))
                es.add((n, m))

    def leaveHandler(self, g, n):
        '''Node leaves one of the edge's compartments, remove any incident edges
        that no longer have the correct orientation.

        :param g: the network
        :param n: the node'''
        es = self.elements()
        for (nn, mm) in g.edges(n):
            match = self.matches(g, nn, mm)
            if match == -1:
                #print('edge ({m}, {n}) leaves {l}'.format(n = nn, m = mm, l = self._name))
                es.discard((mm, nn))
            else:
                if match == 1:
                    #print('edge ({n}, {m}) leaves {l}'.format(n = nn, m = mm, l = self._name))
                    es.discard((nn, mm))

    def enterHandler(self, g, n):
        '''Node enters one of the edge's compartments, add any incident edges
        that now have the correct orientation.

        :param g: the network
        :param n: the node'''
        es = self.elements()
        for (nn, mm) in g.edges(n):
            match = self.matches(g, nn, mm)
            if match == -1:
                #print('edge ({m}, {n}) enters {l}'.format(n = nn, m = mm, l = self._name))
                es.add((mm, nn))
            else:
                if match == 1:
                    #print('edge ({n}, {m}) enters {l}'.format(n = nn, m = mm, l = self._name))
                    es.add((nn, mm))

    def removeHandler(self, g, e):
        '''An edge is removed from the network, check whether it was in this locus and
        remove it if so.

        :param, g: the network
        :param e: the edge'''
        (n, m) = e
        es = self.elements()
        match = self.matches(g, n, m)
        if match == -1:
            #print('edge ({m}, {n}) removed from {l}'.format(n = n, m = m, l = self._name))
            es.discard((m, n))
        else:
            if match == 1:
                #print('edge ({n}, {m}) removed from {l}'.format(n = n, m = m, l = self._name))
                es.discard((n, m))


class CompartmentedModel(Process):
    '''The base class a :term:`compartmented model of disease`, which
    represents a disease as a collection of discrete :term:`compartments` with each
    compartment specifying some facet of the disease's progression. Nodes
    transition between compartments with some probability, so the disease's
    progression forms a stochastic process in which nodes typically move
    through several compartments according to the probabilities defined.

    When run, a compartmented model generates results consisting of the size of each
    compartment at the end of the experiment. This can be changed by overriding the
    :meth:`results` method.'''
    
    # model state variables
    COMPARTMENT = 'compartment'                 #: Node attribute holding the node's compartment.
    OCCUPIED = 'occupied'                       #: Edge attribute, True if infection travelled along the edge.
    T_OCCUPIED = 'occupationTime'               #: Edge attribute holding the time the infection crossed the edge.

    def __init__( self ):
        super(CompartmentedModel, self).__init__()


    # ---------- Setup and initialisation ----------

    def reset( self ):
        '''Reset the model ready to be built.'''
        super(CompartmentedModel, self).reset()
        self._compartments = []
        self._effects = dict()
        self._eventDistribution = None

    def setUp(self, params):
        '''Set up the initial population of nodes into compartments.

        :param params: the simulation parameters'''
        super(CompartmentedModel, self).setUp(params)

        # initialise all nodes to an empty compartment
        # (so we can assume all nodes have a compartment attribute)
        g = self.network()
        for n in g.nodes():
            g.nodes[n][self.COMPARTMENT] = None

        # mark edges as unoccupied
        for (_, _, data) in g.edges(data=True):
            data[self.OCCUPIED] = False

        # place nodes in initial compartments
        self.initialCompartments()

    def initialCompartmentDistribution( self ):
        '''Return the initial distribution of nodes to compartments. The
        result should be a valid distribution, with probabilities summing
        to one. This is used by :meth:`initialCompartments` to set the initial
        compartment of each node.

        :returns: a list of (compartment, probability) pairs'''
        return self._compartments

    def initialCompartments( self ):
        '''Place each node in the network into its initial compartment. The default
        initialises the nodes into a random compartment according to the initial
        compartment distribution returned by :meth:`initialCompartmentDistribution`.
        This method may be overridden to, for example structure the initial
        population non-randomly.'''

        # get the initial compartment distribution
        dist = self.initialCompartmentDistribution()

        # assign nodes to compartments
        g = self.network()
        for n in g.nodes():
            # select a compartment according to the initial distribution
            r = numpy.random.random()
            a = 0.0
            for (c, p) in dist:
                a = a + p
                if r <= a:
                    # change node's compartment
                    self.changeCompartment(n, c)

                    # on to the next node
                    break


    # ---------- Termination and results ----------

    def compartments(self):
        '''Return the set of compartments.

        :returns: the compartments'''
        return self._compartments

    def compartment( self, c ):
        '''Return all the nodes currently in a particular compartment in a network. This works
        for all compartments, not just those that are loci for dynamics.

        :param c: the compartment
        :returns: a collection of nodes'''
        g = self.network()
        return [ n for n in g.nodes() if g.nodes[n][self.COMPARTMENT] == c]

    def results(self):
        '''Create a dict of experimental results for the experiment, consisting of the final
        sizes of all the compartments.

        :returns: a dict of experimental results'''
        rc = super(CompartmentedModel, self).results()

        for (c, _) in self.compartments():
            rc[c] = 0
        g = self.network()
        for n in g.nodes():
            c = g.nodes[n][self.COMPARTMENT]
            rc[c] = rc[c] + 1
        return rc

    def skeletonise(self):
        '''Remove unoccupied edges from the network. This leaves the network
        consisting of only "occupied" edges that were used to transmit the
        infection between nodes, also known as the :term:`contact tree`. Note
        that this process means that further dynamics over the network probably
        don't make sense, unless you're actually wanting to run on the residual
        network post-infection.

        :returns: the network with unoccupied edges removed'''

        # find all unoccupied edges
        g = self.network()
        edges = []
        for (n, m, data) in g.edges(data=True):
            if (self.OCCUPIED not in data.keys()) or (data[self.OCCUPIED] != True):
                # edge is unoccupied, mark it to be removed
                # (safe because there are no parallel edges)
                edges.insert(0, (n, m))

        # remove all the unoccupied edges
        g.remove_edges_from(edges)

        return g


    # ---------- Managing compartments ----------

    def addCompartment( self, c, p = 0.0 ):
        '''Add a compartment to the model. A node is assigned to the compartment
        initially with the given probability. The probabilities for all compartments
        in the model must sum to 1.

        :param c: the compartment name
        :param p: the initial occupancy probability (defaults to  0.0)'''
        self._compartments.append((c, p))

    def trackNodesInCompartment(self, c, name = None):
        '''Add a locus tracking nodes in a given compartment.

        :param c: the compartment to track
        :param name: (optional) the name of the locus (defaults to the compartment name)
        :returns: the locus used to track the nodes'''
        if name is None:
            name = c
        if name in self:
            raise NameError('Locus {c} already exists in model'.format(c = name))
        else:
            # add locus
            locus = CompartmentedNodeLocus(name, c)
            return self.addLocus(name, locus)

    def trackEdgesBetweenCompartments(self, l, r, name = None):
        '''Add a locus to track edges with endpoint nodes in the given compartments.

        :param l: the compartment of the left node
        :param r: the compartment of the right node
        :param name: (optional) the name of the locus (defaults to a combination of the two compartment names)
        :returns: the locus used to track the nodes'''
 
        if name is None:
            name = '{l}-{r}'.format(l = l, r = r)
        if name in self:
            raise NameError('Locus {c} already exists in model'.format(c = name))
        else:
            # add locus
            locus = CompartmentedEdgeLocus(name, l, r)
            return self.addLocus(name, locus)

    def addLocus(self, n, l = None):
        '''Add a locus to the model, initialising the handler functions.

        :param n: the name
        :param l: (optional) the locus
        :returns: the locus'''
        locus = super(CompartmentedModel,self).addLocus(n, l)
        for c in locus.compartments():
            if c not in self._effects.keys():
                self._effects[c] = []
            self._effects[c].append((locus.addHandler, locus.leaveHandler, locus.enterHandler, locus.removeHandler))

    def _addHandlers( self, c, lh, eh ):
        '''Add handler functions for a compartment, to be called whenever the contents
        of that compartment change. This links compartments to loci.

        Handlers are functions that take a network and a node
        and provide the appropriate behaviour for when the node
        enters (or leaves) the compartment.
        
        :param c: the compartment
        :param lh: the leave handler
        :param eh: the enter handler'''
        if c not in self._effects.keys():
            self._effects[c] = [ (lh, eh) ]
        else:
            self._effects[c].append((lh, eh))

    def _handlerCompartments(self, e):
        '''Return the compartments that a given element's change might affect.

        :param e: a node or edge
        :returns: a list of compartments'''
        g = self.network()
        if isinstance(e, tuple):
            # element is an edge, check compartments at the endpoints
            (n, m) = e
            cs = [ g.nodes[n][self.COMPARTMENT], g.nodes[m][self.COMPARTMENT] ]
        else:
            # element is a node, check its own compartmnent
            cs = [ g.nodes[e][self.COMPARTMENT] ]
        return cs

    def _callAddHandlers( self, e ):
        '''Call all handlers affected by a mnode or edge being added to the network.

        :param e the node or edge'''
        g = self.network()
        for c in self._handlerCompartments(e):
            if c in self._effects.keys():
                for (ah, _, _, _) in self._effects[c]:
                    ah(g, e)

    def _callLeaveHandlers( self, e, c ):
        '''Call all handlers affected by a node or edge leaving a compartment.

        :param e: the node or edge
        :param c: the compartment'''
        g = self.network()
        for c in self._handlerCompartments(e):
            if c in self._effects.keys():
                for (_, lh, _, _) in self._effects[c]:
                    lh(g, e)

    def _callEnterHandlers( self, e, c ):
        '''Call all handlers affected by a node or edge entering a compartment.

        :param e: the node or edge
        :param c: the compartment'''
        g = self.network()
        for c in self._handlerCompartments(e):
            if c in self._effects.keys():
                for (_, _, eh, _) in self._effects[c]:
                    eh(g, e)

    def _callRemoveHandlers( self, e ):
        '''Call all handlers affected by a node or edge being removed from the network.

        :param e: the node or edge'''
        g = self.network()
        for c in self._handlerCompartments(e):
            if c in self._effects.keys():
                for (_, _, _, rh) in self._effects[c]:
                    rh(g, e)


    # ---------- Accessing and evolving the network ----------

    def setCompartment(self, n, c):
        '''Set the compartment of a node. This assumes that the node doesn't already have
        a compartment set, and so should be used only for initialising new nodes: in all
        other cases, use :meth:`changeCompartment`.

        :param n: the node
        :param c: the new compartment for the node'''
        g = self.network()

        # set the correct node attribute
        g.nodes[n][self.COMPARTMENT] = c

        # propagate the change to any other compartments
        self._callEnterHandlers(n, c)

    def getCompartment(self, n):
        '''Return the compartment of a node.

        :parak n: the node
        :returns: its compartment'''
        return self.network().nodes[n][self.COMPARTMENT]

    def changeCompartment( self, n, c ):
        '''Change the compartment of a node.

        :param n: the node
        :param c: the new compartment for the node'''
        g = self.network()

        # propagate effects of leaving the current compartment
        self._callLeaveHandlers(n, g.nodes[n][self.COMPARTMENT])

        # record new compartment on node
        g.nodes[n][self.COMPARTMENT] = c

        # propagate effects of entering new compartment
        self._callEnterHandlers(n, c)

    def markOccupied( self, e, t ):
        '''Mark the given edge as having been occupied by the dynamics, i.e., to
        have been traversed in transmitting the disease, at time t.

        :param e: the edge
        :param t: the time at which it was occupied'''
        g = self.network()
        (n, m) = e
        data = g.get_edge_data(n, m)
        data[self.OCCUPIED] = True
        data[self.T_OCCUPIED] = t
        
    def addNode(self, n, compartment, **kwds):
        '''Add a node to the working network, adding it to the appropriate compartment.

        :param n: the new node
        :param compartment: compartment for the node
        :param kwds: (optional) node attributes'''
        super(CompartmentedModel, self).addNode(n, **kwds)
        self.setCompartment(n, compartment)

    def removeNode(self, n):
        '''Remove a node from the working network, updating any affected compartments.

        :param n: the node'''

        # remove node from any compartments
        self._callRemoveHandlers(n)

        # remove the node
        super(CompartmentedModel, self).removeNode(n)

    def addEdge(self, n, m,  **kwds):
        '''Add an edge between nodes, adding the edge to any appropriate compartments.

        :param n: the start node
        :param m: the end node
        :param kwds: (optional) edge attributes'''
        super(CompartmentedModel, self).addEdge(n, m, **kwds)

        # add edge to any compartments it should be in
        self._callAddHandlers((n, m))

    def removeEdge(self, n, m):
        '''Remove an edge from the working network and from any compartments.

        :param n: the start node
        :param m: the end node'''

        # remove edge from any compartments it should be in
        self._callRemoveHandlers((n, m))

        # remove the edge from the network
        super(CompartmentedModel, self).removeEdge(n, m)
